
#ifndef THAECS_H
#define THAECS_H

#include "sofa.h"
#include "sofam.h"
#include <cstring>
#include <stdexcept>
#include <vector>

enum { kPrevSec='s', kNextSec='S', 
       kPrevMin='m', kNextMin='M', 
       kPrevHour='h', kNextHour='H',
       kPrevDay='d', kNextDay='D',
       kPrevWeek='w', kNextWeek='W',
       kPrevYear='y', kNextYear='Y'};

class GenPosVelVec
{
  typedef double array_j[6];
  typedef double array_s[2][3];
  typedef double array_a[3];
  typedef long double longDouble;
public:
  union{
    array_j joined;
    array_s split;
  };
  enum units_l { kKM=1000, kAU=149597870700, kLS=299792458 }; //kilometer, astronomic unit, light secound
  enum units_t { kSec=1, kDay=24*3600 }; //second, day
public:
  GenPosVelVec(){
    for(int i=0; i<6; i++)
      joined[i]=0;
  }
  void convertUnitsPos(units_l fromLenU, units_l toLenU){
    longDouble factor=((longDouble)fromLenU)/((longDouble)toLenU);
    for(int i=0; i<3; i++)
      joined[i]=joined[i]*factor;
  }
  void convertUnitsVel(units_l fromLenU, units_t fromTimeU, units_l toLenU, units_t toTimeU){
    longDouble factor=((longDouble)fromLenU)/((longDouble)toLenU)*((longDouble)toTimeU)/((longDouble)fromTimeU);
    for(int i=3; i<6; i++)
      joined[i]=joined[i]*factor;
  }
  void convertUnitsVel(units_l fromLenU, units_l toLenU){
    //convert position and time units with given position units and guessed time units
    //usual velocity units are namely: km/sec or AU/day or c=LS/sec
    units_t fromTimeU, toTimeU;
    if(fromLenU==kKM ||
       fromLenU==kLS) fromTimeU=kSec;
    if(fromLenU==kAU) fromTimeU=kDay;
    if(toLenU==kKM ||
       toLenU==kLS) toTimeU=kSec;
    if(toLenU==kAU) toTimeU=kDay;
    convertUnitsVel(fromLenU, fromTimeU, toLenU, toTimeU);    
  }
  void convertUnits(units_l fromLenU, units_l toLenU){
    convertUnitsPos(fromLenU, toLenU);
    convertUnitsVel(fromLenU, toLenU);
  }
  double& operator[](int idx)        { return joined[idx]; }
  const double& operator[](int idx) const  { return joined[idx]; }
  array_a& operator()(int idx)       { return split[idx]; }
  const array_a& operator()(int idx) const { return split[idx]; }
};


class Angle_t
{
  double angle;
  
  double fullDegrees() const { return angle*180/DPI; }
  double fullArcMins() const { return angle*180/DPI*60; }
  double fullArcSecs() const { return angle*180/DPI*3600; }

  double fullHours() const { return angle*12/DPI; }
  double fullMins() const { return angle*12/DPI*60; }
  double fullSecs() const { return angle*12/DPI*3600; }
public:
  Angle_t(double ang=0):angle(ang) {}
  
  double& operator()()       { return angle; }
  const double& operator()() const { return angle; }

  int deg() const { return dint(fullDegrees()); }
  int arcM() const { return abs(dint(fullArcMins()-dint(fullDegrees())*60)); }
  double arcS() const { return fabs(fullArcSecs()-dint(fullArcMins())*60); }

  int h() const { return dint(fullHours()); }
  int m() const { return abs(dint(fullMins()-dint(fullHours())*60)); }
  double s() const { return fabs(fullSecs()-dint(fullMins())*60); }
};

void exept(std::string strg) { throw std::runtime_error(strg); }
void exept(int code) { throw std::runtime_error(std::to_string(code)); }

class sofa_jd;
class time_dv {
public:
  union{
    int ymd[4];
    struct{
      int yr;
      int mo;
      int dy;
      int _padding;
    };
  };
  union{
    int hmsf[4];
    struct{
      int hr;
      int mi;
      int sc;
      int fr;
    };
  };
  void toSofaJd(sofa_jd& jd) const;
};

class sofa_jd {
  double jd[2];
public:
  double& operator[](int idx){ return jd[idx]; }
  const double& operator[](int idx) const { return jd[idx]; }
  sofa_jd& operator=(const sofa_jd o){ jd[0]=o.jd[0]; jd[1]=o.jd[1]; return *this; }
  sofa_jd& operator+=(const double v) { jd[1]+=v; return *this; }
  sofa_jd& operator-=(const double v) { jd[1]-=v; return *this; }
  double MJD() const { return (jd[0]-2400000.5)+jd[1]; }
  double JD() const { return jd[0]+jd[1]; }
  void toTime(time_dv& dv, int prec) const;
  double dayFraction() const { return fmod(fmod(jd[0], 1.0) + fmod(jd[1], 1.0), 1.0); }
  friend double operator-(const sofa_jd& a,const sofa_jd& b);
};
double operator-(const sofa_jd& a,const sofa_jd& b){ return (a.jd[0]-b.jd[0])+(a.jd[1]-b.jd[1]); }


class TimeInstant {
public:
  TimeInstant(){ std::memset(this, 0x0, sizeof(*this)); }
  sofa_jd utc;
  sofa_jd ut1;
  sofa_jd tai;
  sofa_jd tt;
  sofa_jd tcg;
  sofa_jd tdb;
  sofa_jd tcb;
  double dut;
  double dat;
  double dt;
  double dtr;
  double ctt;
  double spiceT;
  double era;
  double X;
  double Y;
  double dX;
  double dY;
  double s;
  double xp;
  double yp;
  double sp;
  void utc2ut1() { ut1=utc; ut1[1]+=dut/3600/24; if(!dut)exept("zero dut"); }
  void utc2tai() { tai=utc; tai[1]+=dat/3600/24; if(!dat)exept("zero dat"); }
  void tai2utc() { utc=tai; utc[1]-=dat/3600/24; if(!dat)exept("zero dat"); }
  void tai2tt() { int error=iauTaitt(tai[0], tai[1], &tt[0], &tt[1]); tt[1]+=ctt/1000/1000/24/3600; if(error)exept(error); }
  void tt2tai() { int error=iauTttai(tt[0], tt[1], &tai[0], &tai[1]); tai[1]-=ctt/1000/1000/24/3600; if(error)exept(error); }
  void ut2tt() { int error=iauUt1tt(ut1[0], ut1[1], dt, &tt[0], &tt[1]); if(error)exept(error); }
  void tt2tcg() { int error=iauTttcg(tt[0], tt[1], &tcg[0], &tcg[1]); if(error)exept(error); }
  void tt2tdb() { int error=iauTttdb(tt[0], tt[1], dtr, &tdb[0], &tdb[1]); if(error)exept(error); }
  void tdb2tt() { int error=iauTdbtt(tdb[0], tdb[1], dtr, &tt[0], &tt[1]); if(error)exept(error); }
  void tdb2spice() { spiceT=((tdb[0]-2451545)+tdb[1])*24*3600; }
  void tdb2tcb() { int error=iauTdbtcb(tdb[0], tdb[1], &tcb[0], &tcb[1]); if(error)exept(error); }
  void ut12era() { era=iauEra00(ut1[0], ut1[1]); }
};

struct TerrestrialLocation {
  double lat;
  double lon;
  double elev;
  double temp;
  double press;
  double relhum;
  double wavelength;
  iauASTROM astrom;
};
#endif
